@@ -1,6 +1,4 @@
-import numpy as np 
-
-
+import numpy as np
 
 
 def sin(a):
@@ -21,11 +19,12 @@ def sin(a):
         return funcNode(np.sin, np.cos, None, a, None)
     elif isinstance(a, int) or isinstance(a, float):
         n = valNode()
-        n._set_val(np.sin(a))
+        n.set_val(np.sin(a))
         return n
     else:
         raise TypeError
 
+
 def cos(a):
     """
     Implements cosine to contribute to a computational graph. Converts integers and floats to
@@ -44,11 +43,12 @@ def cos(a):
         return funcNode(np.cos, lambda x: -np.sin(x), None, a, None)
     elif isinstance(a, int) or isinstance(a, float):
         n = valNode()
-        n._set_val(np.cos(a))
+        n.set_val(np.cos(a))
         return n
     else:
         raise TypeError
 
+
 def tan(a):
     """
     Implements tangent to contribute to a computational graph. Converts integers and floats to
@@ -67,11 +67,12 @@ def tan(a):
         return funcNode(np.tan, lambda x: 1 / (np.cos(x)) ** 2, None, a, None)
     elif isinstance(a, int) or isinstance(a, float):
         n = valNode()
-        n._set_val(np.tan(a))
+        n.set_val(np.tan(a))
         return n
     else:
         raise TypeError
 
+
 def sinh(a):
     """
     Implements hyperbolic sine to contribute to a computational graph. Converts integers and floats to
@@ -90,11 +91,12 @@ def sinh(a):
         return funcNode(np.sinh, np.cosh, None, a, None)
     elif isinstance(a, int) or isinstance(a, float):
         n = valNode()
-        n._set_val(np.sinh(a))
+        n.set_val(np.sinh(a))
         return n
     else:
         raise TypeError
-    
+
+
 def cosh(a):
     """
     Implements hyperbolic cosine to contribute to a computational graph. Converts integers and floats to
@@ -113,11 +115,12 @@ def cosh(a):
         return funcNode(np.cosh, np.sinh, None, a, None)
     elif isinstance(a, int) or isinstance(a, float):
         n = valNode()
-        n._set_val(np.cosh(a))
+        n.set_val(np.cosh(a))
         return n
     else:
         raise TypeError
 
+
 def tanh(a):
     """
     Implements hyperbolic tangent to contribute to a computational graph. Converts integers and floats to
@@ -133,10 +136,10 @@ def tanh(a):
         TypeError if value is not of type int, float, or Node.
         """
     if isinstance(a, Node):
-        return funcNode(np.tanh, lambda x: 1-np.tanh(x)**2, None, a, None)
+        return funcNode(np.tanh, lambda x: 1 - np.tanh(x) ** 2, None, a, None)
     elif isinstance(a, int) or isinstance(a, float):
         n = valNode()
-        n._set_val(np.tanh(a))
+        n.set_val(np.tanh(a))
         return n
     else:
         raise TypeError
@@ -157,15 +160,29 @@ def sqrt(a):
         TypeError if value is not of type int, float, or Node.
         """
     if isinstance(a, Node):
-        return funcNode(np.sqrt, lambda x: 0.5/np.sqrt(x), None, a, None)
+        return funcNode(np.sqrt, lambda x: 0.5 / np.sqrt(x), None, a, None)
     elif isinstance(a, int) or isinstance(a, float):
         n = valNode()
-        n._set_val(np.sqrt(a))
+        n.set_val(np.sqrt(a))
         return n
     else:
         raise TypeError
 
+
 def logistic(a):
+    """
+        Implements logistic function to contribute to a computational graph. Converts integers and floats to
+        Node's, and inserts derivatives into the computational graph.
+
+        Args:
+            a (int, float, Node): Value to be calculated through tan.
+
+        Returns:
+            Node with values and derivatives along with a corresponding computational graph.
+
+        Raises:
+            TypeError if value is not of type int, float, or Node.
+            """
     if isinstance(a, Node):
         return funcNode(_logistic, lambda x: _logistic(x) * (1 - _logistic(x)), None, a, None)
     elif isinstance(a, int) or isinstance(a, float):
@@ -177,8 +194,15 @@ def logistic(a):
 
 
 def _logistic(a):
-    return (np.tanh(a / 2) + 1) * 0.5
+    """
+    Helper function for logistic, implements logistic function for int or float data type.
+    Args:
+        a (int, float, Node): Real number whose logistic function is to be calculated.
 
+    Returns:
+        Logistic value of the argument a.
+    """
+    return (np.tanh(a / 2) + 1) * 0.5
 
 
 def exp(a):
@@ -199,11 +223,12 @@ def exp(a):
         return funcNode(np.exp, np.exp, None, a, None)
     elif isinstance(a, int) or isinstance(a, float):
         n = valNode()
-        n._set_val(np.exp(a))
+        n.set_val(np.exp(a))
         return n
     else:
         raise TypeError
 
+
 def log(a):
     """
     Alias for ln(a).
@@ -216,6 +241,8 @@ def log(a):
     """
 
     return ln(a)
+
+
 def ln(a):
     """
     Implements the natural logarithm to calculate values, derivative, and to contribute to a computational graph. Converts integers and floats to
@@ -231,16 +258,27 @@ def ln(a):
         return funcNode(np.log, lambda x: 1 / x, None, a, None)
     elif isinstance(a, int) or isinstance(a, float):
         n = valNode()
-        n._set_val(np.log(a))
+        n.set_val(np.log(a))
         return n
     else:
         raise TypeError
 
+
 def _log_ab(arg, base=np.exp(1)):
     return np.log(arg) / np.log(base)
 
 
 def log_ab(a, b):
+    """
+        Implements the logarithm with arbitrary base b for argument a to calculate values, derivative, and to contribute
+        to a computational graph. Converts integers and floats to Node's, and inserts derivatives into the computational graph.
+
+        Args:
+            a (int, float, Node): The Argument of logarithm.
+            b (int, float, Node): The base of the logarithma
+        Returns:
+            Node with values and derivatives along with a corresponding computational graph.
+        """
     if isinstance(a, Node) and isinstance(b, Node):
         return funcNode(_log_ab, lambda x, y: (1 / np.log(y)) * 1 / x,
                         lambda x, y: (-(1 / y) * np.log(x) * ((1 / np.log(y)) ** 2)),
@@ -299,7 +337,7 @@ class Node:
             r = other
         elif isinstance(other, int) or isinstance(other, float):
             r = valNode()
-            r._set_val(other)
+            r.set_val(other)
         else:
             raise TypeError
         return funcNode(np.add, lambda x, y: 1, lambda x, y: 1, self, r)
@@ -325,7 +363,7 @@ class Node:
             r = other
         elif isinstance(other, int) or isinstance(other, float):
             r = valNode()
-            r._set_val(other)
+            r.set_val(other)
         else:
             raise TypeError
         return funcNode(np.multiply, lambda x, y: y, lambda x, y: x, self, r)
@@ -350,7 +388,7 @@ class Node:
             r = other
         elif isinstance(other, int) or isinstance(other, float):
             r = valNode()
-            r._set_val(other)
+            r.set_val(other)
         else:
             raise TypeError
         return funcNode(np.subtract, lambda x, y: 1, lambda x, y: -1, self, r)
@@ -369,7 +407,7 @@ class Node:
             r = other
         elif isinstance(other, int) or isinstance(other, float):
             r = valNode()
-            r._set_val(other)
+            r.set_val(other)
         else:
             raise TypeError
         return funcNode(np.divide, lambda x, y: 1 / y, lambda x, y: -x / y / y, self, r)
@@ -377,7 +415,7 @@ class Node:
     def __rtruediv__(self, other):
         if isinstance(other, int) or isinstance(other, float):
             l = valNode()
-            l._set_val(other)
+            l.set_val(other)
         else:
             raise TypeError
         return funcNode(np.divide, lambda x, y: 1 / y, lambda x, y: -x / y / y, l, self)
@@ -393,15 +431,15 @@ class Node:
             r = other
         elif isinstance(other, int) or isinstance(other, float):
             r = valNode()
-            r._set_val(other)
+            r.set_val(other)
         else:
-            raise TypeError
+            raise TypeError("unsupported input type(s) for the operand")
         return funcNode(np.power, lambda x, y: y * np.power(x, y - 1), lambda x, y: np.power(x, y) * np.log(x), self, r)
 
     def __rpow__(self, other):
         if isinstance(other, int) or isinstance(other, float):
             l = valNode()
-            l._set_val(other)
+            l.set_val(other)
         else:
             raise TypeError
         return funcNode(np.power, lambda x, y: y * np.power(x, y - 1), lambda x, y: np.power(x, y) * np.log(x), l, self)
@@ -412,14 +450,15 @@ class Node:
         """
         return self
 
-        
+
 class valNode(Node):
     """
     Sets variables as valNode class, which allows the users to auto-differentiate functions that use them. valNode class
     overrides native python functions and unary operators, in order to construct a sequential computational graph. Sets
     the initial derivative (0) for each assigned variable.
     """
-    def __init__(self, name = None):
+
+    def __init__(self, name=None):
         super().__init__()
         self.der = 0
         self.name = name
@@ -435,6 +474,16 @@ class valNode(Node):
         """
         self.val = val
 
+    def set_val(self, val):
+        """
+        Alias for _set_val(self, val).  Sets the numeric value for a valNode.
+
+        Args:
+            val(int, float): numeric value of the valNode.
+
+        """
+        self._set_val(val)
+
     def __str__(self):
         """
         Prints information about a specified valNode.
@@ -455,16 +504,19 @@ class valNode(Node):
         if self.name is None:
             return self.val, {self.name: 0}
         return self.val, {self.name: 1}
-    
+
     def forward_pass(self):
         self.der = 0
         return self.val
+        
+    def reverse_pass(self, partial, adjoint):
+        if self.name == None:
+            return dict()
+        self.der += partial*adjoint
+        return {self.name: self}
 
-    def reverse(self, partial, adjoint):
-        # print(partial, adjoint)
-        if self.name != None:
-            self.der += partial * adjoint
-
+    def _reset_der(self):
+        self.der = 0
 
 class funcNode(Node):
     def __init__(self, func, leftdf, rightdf, left, right):
@@ -509,47 +561,135 @@ class funcNode(Node):
         else:
             self.val = self.func(self.left.forward_pass())
         return self.val
-
-    def reverse(self, partial, adjoint):
-        if self.right != None:
+    
+    def reverse_pass(self, partial, adjoint):
+        if self.right != None: 
             lder = self.leftdf(self.left.val, self.right.val)
             rder = self.rightdf(self.left.val, self.right.val)
-            self.left.reverse(lder, partial * adjoint)
-            self.right.reverse(rder, partial * adjoint)
+            lvars = self.left.reverse_pass(lder, partial*adjoint)
+            rvars = self.right.reverse_pass(rder, partial*adjoint)
+            variables = dict()
+            for v in rvars:
+                if v not in lvars:
+                    lvars[v] = rvars[v]
         else:
             lder = self.leftdf(self.left.val)
-            self.left.reverse(lder, partial * adjoint)
-
+            lvars = self.left.reverse_pass(lder, partial*adjoint) 
+            
+        return lvars
         # return self.val
+    
+    def reverse(self):
+        self.forward_pass()
+        variables = self.reverse_pass(1,1)
+        der = { v: node.der for v, node in variables.items() }
+        return self.val, der
+    
+    
+class vector:
+    def __init__(self, *args):
+        self.size = len(args)
+        self.elements = args
+    
+    def set_val(self, array):
+        if len(array) != self.size:
+            raise ValueError(f"Input size has a mismatch with the vector size ({self.size})")
+        for node, val in zip(self.elements, array):
+            # print(val)
+            node.set_val(val)
+
+    def __getitem__(self, key):
+        return self.elements[key]
+        
+    def __iter__(self, key):
+        return iter(self.elements)
+    
+    def grad(self, var):
+        '''
+            auto-diff for vector functions
+            currently we only use reverse mode to calculate gradient for vector functions
+            input: the variable or vector variables var, which is used to
+             define the function and for which we need to calculate the derivatives
+            return: Jacobian matrix
+        '''
+        der_array = []
+        var._reset_der()
+        for f in self.elements:
+            f.forward_pass()
+            f.reverse_pass(1,1)
+            der_array.append(var.der)
+            var._reset_der()
+        return der_array
+    
+    def evaluate(self):
+        '''
+            evaluation for vector functions
+        '''
+        return [ f.forward_pass() for f in self.elements ]        
+    
+    @property
+    def der(self):
+        return [ node.der for node in self.elements ]
+    
+    def _reset_der(self):
+        '''
+           this function only works if it is a variable vector
+        '''
+        for node in self.elements:
+            node.der = 0
+    
+def variables(name, size = None):
+    if size == None or size == 1:
+        return valNode(name)
+    elems = [ valNode(name+'___'+str(i)) for i in range(size) ]
+    return vector(*elems)
+
+    
 
             
-            
+
 if __name__ == '__main__':
-
-    
     # v = vector([1,2,3])
+
+    # example to define scalar function with multiple input variables
+
+    x = variables('x')
+    y = variables('y')
+    f = sin(ln(x))+tan(x*x+y*x+x*x*x*y)
+    x.set_val(2)
+    y.set_val(3)
+    print(f.forward())
+    print(f.reverse())
     
-    def f(v):
+    # example to define vector function with vectorized input variables
+    
+    v = variables('v', 2)
+    v.set_val([2,3])    
+    print(v.evaluate())
+    print(v.grad(v))
+    def func(v):
         '''
             f takes a size=3 vector and output a size=2 vector
         '''
-        f1 = sin(ln(v[0]))+tan(v[0]**2+v[0]*v[1]+v[2])
-        f2 = ln(sin(exp(v[0])+v[1])) + exp(v[1]) + v[2]
-        return f1,f2
-
-    '''x = valNode('x')
-    y = valNode('y')
-    c = valNode('c')
-    f = sin(log(x)) + tan(x * x + y * x + x ** 3 * y)
-    x._set_val(2)
-    y._set_val(3)
-    c._set_val(np.pi)
-    print(f.forward())
-    f.forward_pass()
-    f.reverse(1,1)
-    print(f.val, x.der, y.der)
+        f1 = sin(ln(v[0]))+tan(v[0]**2+v[0]*v[1]+v[0]**3*v[1])
+        f2 = v[0]*v[1]+v[1]**2+sqrt(v[0])
+        return vector(f1,f2)
+    f = func(v)
+    print(f.evaluate())
+    print(f.grad(v))
+        
+        
+    # vector function can also take a scalar input variable
+    def func1(x):
+        f1 = x**2
+        f2 = x**3
+        return vector(f1, f2)
     
-    x.der = 0
+    f = func1(x)
+    print(f.evaluate())
+    print(f.grad(x))
+        
+    '''x.der = 0
     y.der = 0
     f = log(sin(exp(x)+y)) + exp(y) + x
     x._set_val(0)
@@ -630,7 +770,3 @@ if __name__ == '__main__':
 
     for var in actual_f_grad.keys():
         assert np.isclose(actual_f_grad[var], reverse_grads[var])'''
-
-
-
-
